# -*- coding: utf-8 -*-
from django.conf.urls import include, url
from django.utils.module_loading import import_string


def default_urlpatterns(provider):
    package = provider.get_package()

    try:
        login_view = import_string(package + '.views.login')
    except ImportError:
        raise ImportError(
            "The login view for the '{id}' provider is lacking from the "
            "'views' module of its app.\n"
            "You may want to add:\n"
            "from allauth_cas.views import CASLoginView\n\n"
            "login = CASLoginView.adapter_view(<LocalCASAdapter>)"
            .format(id=provider.id)
        )

    try:
        callback_view = import_string(package + '.views.callback')
    except ImportError:
        raise ImportError(
            "The callback view for the '{id}' provider is lacking from the "
            "'views' module of its app.\n"
            "You may want to add:\n"
            "from allauth_cas.views import CASCallbackView\n\n"
            "callback = CASCallbackView.adapter_view(<LocalCASAdapter>)"
            .format(id=provider.id)
        )

    try:
        logout_view = import_string(package + '.views.logout')
    except ImportError:
        logout_view = None

    urlpatterns = [
        url('^login/$', login_view,
            name=provider.id + '_login'),
        url('^login/callback/$', callback_view,
            name=provider.id + '_callback'),
    ]

    if logout_view is not None:
        urlpatterns += [
            url('^logout/$', logout_view,
                name=provider.id + '_logout'),
        ]

    return [url('^' + provider.get_slug() + '/', include(urlpatterns))]



# -*- coding: utf-8 -*-
from django.contrib.auth.signals import user_logged_out
from django.dispatch import receiver

from allauth.account.adapter import get_adapter
from allauth.account.utils import get_next_redirect_url
from allauth.socialaccount import providers

from . import CAS_PROVIDER_SESSION_KEY


@receiver(user_logged_out)
def cas_account_logout(sender, request, **kwargs):
    provider_id = request.session.get(CAS_PROVIDER_SESSION_KEY)

    if not provider_id:
        return

    provider = providers.registry.by_id(provider_id, request)

    if not provider.message_suggest_caslogout_on_logout(request):
        return

    next_page = (
        get_next_redirect_url(request) or
        get_adapter(request).get_logout_redirect_url(request)
    )

    provider.add_message_suggest_caslogout(
        request, next_page=next_page,
        level=provider.message_suggest_caslogout_on_logout_level(request),
    )



# -*- coding: utf-8 -*-
from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class CASAccountConfig(AppConfig):
    name = 'allauth_cas'
    verbose_name = _("CAS Accounts")

    def ready(self):
        from . import signals  # noqa



# -*- coding: utf-8 -*-
from allauth.socialaccount.providers.base import ProviderAccount

from allauth_cas.providers import CASProvider


class ExampleCASAccount(ProviderAccount):
    pass


class ExampleCASProvider(CASProvider):
    id = 'theid'
    name = 'The Provider'
    account_class = ExampleCASAccount


provider_classes = [ExampleCASProvider]



# -*- coding: utf-8 -*-
from six.moves.urllib.parse import parse_qsl

import django
from django.contrib import messages
from django.template.loader import render_to_string
from django.utils.http import urlencode
from django.utils.safestring import mark_safe

from allauth.socialaccount.providers.base import Provider

if django.VERSION >= (1, 10):
    from django.urls import reverse
else:
    from django.core.urlresolvers import reverse


class CASProvider(Provider):

    def get_auth_params(self, request, action):
        settings = self.get_settings()
        ret = dict(settings.get('AUTH_PARAMS', {}))
        dynamic_auth_params = request.GET.get('auth_params')
        if dynamic_auth_params:
            ret.update(dict(parse_qsl(dynamic_auth_params)))
        return ret

    ##
    # Data extraction from CAS responses.
    ##

    def extract_uid(self, data):
        """Extract the user uid.

        Notes:
            Each pair ``(provider_id, uid)`` is unique and related to a single
            user.

        Args:
            data (uid (str), extra (dict)): CAS response. Example:
                ``('alice', {'name': 'Alice'})``

        Returns:
            str: Default to ``data[0]``, user identifier for the CAS server.

        """
        uid, _ = data
        return uid

    def extract_common_fields(self, data):
        """Extract the data to pass to `SOCIALACCOUNT_ADAPTER.populate_user()`.

        Args:
            data (uid (str), extra (dict)): CAS response. Example:
                ``('alice', {'name': 'Alice'})``

        Returns:
            dict: Default::

                {
                    'username': extra.get('username', uid),
                    'email': extra.get('email'),
                    'first_name': extra.get('first_name'),
                    'last_name': extra.get('last_name'),
                    'name': extra.get('name'),
                }

        """
        uid, extra = data
        return {
            'username': extra.get('username', uid),
            'email': extra.get('email'),
            'first_name': extra.get('first_name'),
            'last_name': extra.get('last_name'),
            'name': extra.get('name'),
        }

    def extract_email_addresses(self, data):
        """Extract the email addresses.

        Args:
            data (uid (str), extra (dict)): CAS response. Example:
                ``('alice', {'name': 'Alice'})``

        Returns:
            `list` of `EmailAddress`: By default, ``[]``.

            Example::

                [
                    EmailAddress(
                        email='user@domain.net',
                        verified=True, primary=True,
                    ),
                    EmailAddress(
                        email='alias@domain.net',
                        verified=True, primary=False,
                    ),
                ]

        """
        return super(CASProvider, self).extract_email_addresses(data)

    def extract_extra_data(self, data):
        """Extract the data to save to `SocialAccount.extra_data`.

        Args:
            data (uid (str), extra (dict)): CAS response. Example:
                ``('alice', {'name': 'Alice'})``

        Returns:
            dict: By default, ``data``.
        """
        uid, extra = data
        return dict(extra, uid=uid)

    ##
    # Message to suggest users to logout of the CAS server.
    ##

    def add_message_suggest_caslogout(
        self, request, next_page=None, level=None,
    ):
        """Add a message with a link for the user to logout of the CAS server.

        It uses the template ``socialaccount/messages/suggest_caslogout.html``,
        with the ``provider`` and the ``logout_url`` as context.

        Args:
            request: The request to which the message is added.
            next_page (optional): Added to the logout link for the CAS server
                to redirect the user to this url.
                Default: ``request.get_full_path()``
            level: The message level. Default: ``messages.INFO``

        """
        if next_page is None:
            next_page = request.get_full_path()
        if level is None:
            level = messages.INFO

        logout_url = self.get_logout_url(request, next=next_page)

        # DefaultAccountAdapter.add_message is unusable because it always
        # escape the message content.

        template = 'socialaccount/messages/suggest_caslogout.html'
        context = {
            'provider': self,
            'logout_url': logout_url,
        }

        messages.add_message(
            request, level,
            mark_safe(render_to_string(template, context).strip()),
            fail_silently=True,
        )

    def message_suggest_caslogout_on_logout(self, request):
        """Indicates whether the logout message should be sent on user logout.

        By default, it returns
        ``settings.SOCIALACCOUNT_PROVIDERS[self.id]['MESSAGE_SUGGEST_CASLOGOUT_ON_LOGOUT']``
        or ``False``.

        Notes:
            The ``request`` argument is the one trigerring the emission of the
            signal ``user_logged_out``.

        """
        return (
            self.get_settings()
            .get('MESSAGE_SUGGEST_CASLOGOUT_ON_LOGOUT', False)
        )

    def message_suggest_caslogout_on_logout_level(self, request):
        """Level of the logout message issued on user logout.

        By default, it returns
        ``settings.SOCIALACCOUNT_PROVIDERS[self.id]['MESSAGE_SUGGEST_CASLOGOUT_ON_LOGOUT_LEVEL']``
        or ``messages.INFO``.

        Notes:
            The ``request`` argument is the one trigerring the emission of the
            signal ``user_logged_out``.

        """
        return (
            self.get_settings()
            .get('MESSAGE_SUGGEST_CASLOGOUT_ON_LOGOUT_LEVEL', messages.INFO)
        )

    ##
    # Shortcuts functions.
    ##

    def get_login_url(self, request, **kwargs):
        url = reverse(self.id + '_login')
        if kwargs:
            url += '?' + urlencode(kwargs)
        return url

    def get_callback_url(self, request, **kwargs):
        url = reverse(self.id + '_callback')
        if kwargs:
            url += '?' + urlencode(kwargs)
        return url

    def get_logout_url(self, request, **kwargs):
        url = reverse(self.id + '_logout')
        if kwargs:
            url += '?' + urlencode(kwargs)
        return url



# -*- coding: utf-8 -*-
from django.http import HttpResponseRedirect
from django.utils.functional import cached_property

from allauth.account.adapter import get_adapter
from allauth.account.utils import get_next_redirect_url
from allauth.socialaccount import providers
from allauth.socialaccount.helpers import (
    complete_social_login, render_authentication_error,
)
from allauth.socialaccount.models import SocialLogin

import cas

from . import CAS_PROVIDER_SESSION_KEY
from .exceptions import CASAuthenticationError


class AuthAction(object):
    AUTHENTICATE = 'authenticate'
    REAUTHENTICATE = 'reauthenticate'
    DEAUTHENTICATE = 'deauthenticate'


class CASAdapter(object):
    #: CAS server url.
    url = None
    #: CAS server version.
    #: Choices: ``1`` or ``'1'``, ``2`` or ``'2'``, ``3`` or ``'3'``,
    #: ``'CAS_2_SAML_1_0'``
    version = None

    def __init__(self, request):
        self.request = request

    @cached_property
    def renew(self):
        """Controls presence of ``renew`` in requests to the CAS server.

        If ``True``, opt out single sign-on (SSO) functionality of the CAS
        server. So that, user is always prompted for his username and password.

        If ``False``, the CAS server does not prompt users for their
        credentials if a SSO exists.

        The default allows user to connect via an already used CAS server
        with other credentials.

        Returns:
            ``True`` if logged in user has already connected to Django using
            **any** CAS provider in the current session, ``False`` otherwise.

        """
        return CAS_PROVIDER_SESSION_KEY in self.request.session

    @cached_property
    def provider(self):
        """
        Returns a provider instance for the current request.
        """
        return providers.registry.by_id(self.provider_id, self.request)

    def complete_login(self, request, response):
        """
        Executed by the callback view after successful authentication on the
        CAS server.

        Args:
            request
            response (`dict`): Data returned by the CAS server.
                ``response[username]`` contains the user identifier for the
                server, and may contain extra user-attributes.

        Returns:
            `SocialLogin()` object: State of the login-session.

        """
        login = self.provider.sociallogin_from_response(request, response)
        return login

    def get_service_url(self, request):
        """The service url, used by the CAS client.

        According to the CAS spec, the service url is passed by the CAS client
        at several times. It must be the same for all interactions with the CAS
        server.

        It is used as redirection from the CAS server after a succssful
        authentication. So, the callback url is used as service url.

        If present, the GET param ``next`` is added to the service url.
        """
        redirect_to = get_next_redirect_url(request)

        callback_kwargs = {'next': redirect_to} if redirect_to else {}
        callback_url = (
            self.provider.get_callback_url(request, **callback_kwargs))

        service_url = request.build_absolute_uri(callback_url)

        return service_url


class CASView(object):
    """
    Base class for CAS views.
    """
    @classmethod
    def adapter_view(cls, adapter):
        """Transform the view class into a view function.

        Similar to the Django ``as_view()`` method.

        Notes:
            An (human) error page is rendered if any ``CASAuthenticationError``
            is catched.

        Args:
            adapter (:class:`CASAdapter`): Provide specifics of a CAS server.

        Returns:
            A view function. The given adapter and related provider are
            accessible as attributes from the view class.


        """
        def view(request, *args, **kwargs):
            # Prepare the func-view.
            self = cls()

            self.request = request
            self.args = args
            self.kwargs = kwargs

            # Setup and store adapter as view attribute.
            self.adapter = adapter(request)
            self.provider = self.adapter.provider

            try:
                return self.dispatch(request, *args, **kwargs)
            except CASAuthenticationError:
                return self.render_error()

        return view

    def get_client(self, request, action=AuthAction.AUTHENTICATE):
        """
        Returns the CAS client to interact with the CAS server.
        """
        auth_params = self.provider.get_auth_params(request, action)

        service_url = self.adapter.get_service_url(request)

        client = cas.CASClient(
            service_url=service_url,
            server_url=self.adapter.url,
            version=self.adapter.version,
            renew=self.adapter.renew,
            extra_login_params=auth_params,
        )

        return client

    def render_error(self):
        """
        Returns an HTTP response in case an authentication failure happens.
        """
        return render_authentication_error(self.request, self.provider.id)


class CASLoginView(CASView):

    def dispatch(self, request):
        """
        Redirects to the CAS server login page.
        """
        action = request.GET.get('action', AuthAction.AUTHENTICATE)
        SocialLogin.stash_state(request)
        client = self.get_client(request, action=action)
        return HttpResponseRedirect(client.get_login_url())


class CASCallbackView(CASView):

    def dispatch(self, request):
        """
        The CAS server redirects the user to this view after a successful
        authentication.

        On redirect, CAS server should add a ticket whose validity is verified
        here. If ticket is valid, CAS server may also return extra attributes
        about user.
        """
        client = self.get_client(request)

        # CAS server should let a ticket.
        try:
            ticket = request.GET['ticket']
        except KeyError:
            raise CASAuthenticationError(
                "CAS server didn't respond with a ticket."
            )

        # Check ticket validity.
        # Response format on:
        # - success: username, attributes, pgtiou
        # - error: None, {}, None
        response = client.verify_ticket(ticket)

        uid, extra, _ = response

        if not uid:
            raise CASAuthenticationError(
                "CAS server doesn't validate the ticket."
            )

        # Keep tracks of the last used CAS provider.
        request.session[CAS_PROVIDER_SESSION_KEY] = self.provider.id

        data = (uid, extra or {})

        # Finish the login flow.
        login = self.adapter.complete_login(request, data)
        login.state = SocialLogin.unstash_state(request)
        return complete_social_login(request, login)


class CASLogoutView(CASView):

    def dispatch(self, request, next_page=None):
        """
        Redirects to the CAS server logout page.

        next_page is used to let the CAS server send back the user. If empty,
        the redirect url is built on request data.
        """
        action = AuthAction.DEAUTHENTICATE

        redirect_url = next_page or self.get_redirect_url()
        redirect_to = request.build_absolute_uri(redirect_url)

        client = self.get_client(request, action=action)

        return HttpResponseRedirect(client.get_logout_url(redirect_to))

    def get_redirect_url(self):
        """
        Returns the url to redirect after logout.
        """
        request = self.request
        return (
            get_next_redirect_url(request) or
            get_adapter(request).get_logout_redirect_url(request)
        )



# -*- coding: utf-8 -*-
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

import django
from django.conf import settings
from django.test import TestCase

import cas

from allauth_cas import CAS_PROVIDER_SESSION_KEY

if django.VERSION >= (1, 10):
    from django.urls import reverse
else:
    from django.core.urlresolvers import reverse


class CASTestCase(TestCase):

    def client_cas_login(
            self,
            client, provider_id='theid',
            username=None, attributes={}):
        """
        Authenticate client through provider_id.

        Returns the response of the callback view.

        username and attributes control the CAS server response when ticket is
        checked.
        """
        client.get(reverse('{id}_login'.format(id=provider_id)))
        self.patch_cas_response(
            valid_ticket='__all__',
            username=username, attributes=attributes,
        )
        callback_url = reverse('{id}_callback'.format(id=provider_id))
        r = client.get(callback_url, {'ticket': 'fake-ticket'})
        self.patch_cas_response_stop()
        return r

    def patch_cas_response(
            self,
            valid_ticket,
            username=None, attributes={}):
        """
        Patch the CASClient class used by views of CAS providers.

        Arguments determines the response of verify_ticket method:

        - If ticket given as paramater to this method is equal to valid_ticket,
          its return value corresponds to a successful authentication on CAS
          server for user whose login is username argument (default:
          "username") and extra attributes (provided by the server) are
          attributes argument (default: {}).

        - If ticket doesn't match valid_ticket, the response corresponds to a
          reject from CAS server.

        Special values for valid_ticket:

        - If valid_ticket is '__all__', a success response is always returned.
        - If valid_ticket is None, a failure response is always returned.

        Note that valid_ticket sould be a string (which is the type of the
        ticket retrieved from GET parameter on request on the callback view).
        """
        if hasattr(self, '_patch_cas_client'):
            self.patch_cas_response_stop()

        class MockCASClient(object):
            _username = username

            def __new__(self_client, *args, **kwargs):
                version = kwargs.pop('version')
                if version in (1, '1'):
                    client_class = cas.CASClientV1
                elif version in (2, '2'):
                    client_class = cas.CASClientV2
                elif version in (3, '3'):
                    client_class = cas.CASClientV3
                elif version == 'CAS_2_SAML_1_0':
                    client_class = cas.CASClientWithSAMLV1
                else:
                    raise ValueError('Unsupported CAS_VERSION %r' % version)

                client_class._username = self_client._username

                def verify_ticket(self, ticket):
                    if valid_ticket == '__all__' or ticket == valid_ticket:
                        username = self._username or 'username'
                        return username, attributes, None
                    return None, {}, None

                patcher = patch.object(
                    client_class, 'verify_ticket',
                    new=verify_ticket,
                )
                patcher.start()

                return client_class(*args, **kwargs)

        self._patch_cas_client = patch(
            'allauth_cas.views.cas.CASClient',
            MockCASClient,
        )
        self._patch_cas_client.start()

    def patch_cas_response_stop(self):
        self._patch_cas_client.stop()
        del self._patch_cas_client

    def tearDown(self):
        if hasattr(self, '_patch_cas_client'):
            self.patch_cas_response_stop()


class CASViewTestCase(CASTestCase):

    def assertLoginSuccess(self, response, redirect_to=None):
        """
        Asserts response corresponds to a successful login.

        To check this, the response should redirect to redirect_to (default to
        /accounts/profile/, the default redirect after a successful login).
        Also CAS_PROVIDER_SESSION_KEY should be set in the client' session. By
        default, self.client is used.
        """
        if redirect_to is None:
            redirect_to = settings.LOGIN_REDIRECT_URL

        self.assertRedirects(
            response, redirect_to,
            fetch_redirect_response=False,
        )
        self.assertIn(
            CAS_PROVIDER_SESSION_KEY,
            response.wsgi_request.session,
        )

    def assertLoginFailure(self, response):
        """
        Asserts response corresponds to a failed login.
        """
        return self.assertInHTML(
            '<h1>Social Network Login Failure</h1>',
            str(response.content),
        )



# -*- coding: utf-8 -*-
from allauth_cas import views

from .provider import ExampleCASProvider


class ExampleCASAdapter(views.CASAdapter):
    provider_id = ExampleCASProvider.id
    url = 'https://server.cas'
    version = 2


login = views.CASLoginView.adapter_view(ExampleCASAdapter)
callback = views.CASCallbackView.adapter_view(ExampleCASAdapter)
logout = views.CASLogoutView.adapter_view(ExampleCASAdapter)
